/*
 * cmd.c
 * functions for cancan's built-in #commands
 *
 * (created: Finn Arne Gangstad, Dec 25th, 1993)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef AIX
#include <sys/select.h>
#endif

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"
#include "beam.h"

FILE *popen();
int pclose();
int select();

static void show_aliases();
static void show_actions();
static void show_keybinds();
static char *keyfunct_name();
static void keybind_modify();
static int key_collision();
static char *read_seq();
static int get_one_char();
static void show_marks();
static int parse_attributes();
#ifndef OMIT_SCHEDULE
static void show_schedule();
static void parse_schedule();
static long parse_delay();
#endif
static void show_variables();
static int toggle();

static void cmd_action(), cmd_alias(), cmd_bind(), cmd_cancel();
static void cmd_capture(), cmd_echo(), cmd_edit(), cmd_for();
static void cmd_help(), cmd_hilite(), cmd_identify(), cmd_init();
static void cmd_map(), cmd_mark(), cmd_option(), cmd_prefix();
static void cmd_quote(), cmd_read(), cmd_record(), cmd_retrace();
static void cmd_set(), cmd_speedwalk(), cmd_ver(), cmd_wrap();
#ifndef OMIT_SESSIONS
static void cmd_connect(), cmd_snoop(), cmd_zap();
#endif
#ifndef OMIT_SCHEDULE
static void cmd_schedule();
#endif
#ifndef NOSHELL
static void cmd_shell();
#endif
#ifdef TELNETBUG
static void cmd_stdcolour();
#endif

const cmdstruct commands[] = {
   { "help", "[topic]\t\t\thelp on commands, or this text", cmd_help },
   { "17", "command\t\t\trepeat 'command' 17 times", 0 },
#ifndef NOSHELL
   { "!", "shell-command\t\texecute a shell command using /bin/sh", cmd_shell },
   { "?", "shell-command;...[;...]\tif .. then .. else", 0 },
#endif
   { "action", "[name][\"pattern\"][=cmd]\tlist/delete/define actions",
	cmd_action },
   { "alias", "[name[=text]]\t\tlist/delete/define aliases", cmd_alias },
   { "bind", "[name[=command]]\t\tlist/delete/define key bindings", cmd_bind },
   { "cancel", "[number]\t\tcancel editing session", cmd_cancel },
   { "capture", "[filename]\t\tbegin/end of capture to file", cmd_capture },
#ifndef OMIT_SESSIONS
   { "connect", "[id[ init][=address]]\tconnect a new session", cmd_connect },
#endif
   { "echo", "[-n ][text]\t\tdisplay text on screen", cmd_echo },
   { "edit", "\t\t\t\tlist editing sessions", cmd_edit },
   { "for", "a b [c] cmd|var cmd\titerate a command", cmd_for },
   { "hilite", "[attr]\t\t\thighlight your input line", cmd_hilite },
   { "identify", "[startact endact]\tsend MUME client identification",
	cmd_identify },
   { "init", "[-][[+]string]\t\tsend string to host on connect", cmd_init },
   { "map", "[[-]number]\t\tshow/clear (auto)map", cmd_map },
   { "mark", "[string[=attr]]\t\tlist/define markers", cmd_mark },
   { "option", "[[+|-]option ...]\tset your preferences", cmd_option },
   { "prefix", "string\t\t\tprefix all lines with string", cmd_prefix },
   { "quote", "[on|off]\t\t\ttoggle verbatim-flag", cmd_quote },
   { "read", "file\t\t\tread lines from file/pipe and send", cmd_read },
   { "record", "[name]\t\t\trecord typed lines into alias", cmd_record },
   { "retrace", "[[-]number]\t\tretrace the last number steps", cmd_retrace },
#ifndef OMIT_SCHEDULE
   { "schedule", "name[ delay][=value]\tlist/delete/define delayed commands",
	cmd_schedule },
#endif
   { "set", "name[=value]\t\tset variable value", cmd_set },
#ifndef OMIT_SESSIONS
   { "snoop", "id\t\t\ttoggle output display for session", cmd_snoop },
#endif
   { "speedwalk", "[on|off]\t\ttoggle speedwalk mode", cmd_speedwalk },
#ifdef TELNETBUG
   { "stdcolour", "attr\t\t\tset standard background colour", cmd_stdcolour },
#endif
   { "ver", "\t\t\t\tshow cancan version", cmd_ver },
   { "wrap", "[on|off]\t\t\tturn word wrapping on/off", cmd_wrap },
#ifndef OMIT_SESSIONS
   { "zap", "id\t\t\t\tdisconnect a session", cmd_zap },
#endif
   { 0, 0, 0 }
};

static char * const option_names[] = OPTION_NAMES;
static char * const attr_names[] = ATTR_NAMES;
#define get_attr_token(p) ((attr_token)get_token(p, attr_names))

char *recording_name;		/* name of recording alias or NULL */

cmd_func *cmdfunc_by_name(str)
char *str;
{
    register const cmdstruct *p = commands;
    register char *s, *t;
    if(*str) do {
	t = p->name;  s = str;
	if(*s != *t) continue;
	do s++, t++; while(*s && *s == *t);
	if(*s == *t) break;
    } while((++p)->name);
    return p->func;
}

static void cmd_action(arg)
char *arg;
{
    if(!*arg)
	show_actions();
    else
	parse_action(arg);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p = actions;
    if(p) {
	sprintf(insertbuf, "#The following action%s defined:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#action %c%s \"%s\"=%s",
		    (p->active < 0) ? '-' : '+',
		    p->label, p->pattern, p->commands);
    } else
	printf("#No actions are defined.\n");
}

/*
 * parse the #action command
 */
void parse_action(str)
char *str;
{
    actionnode **np;
    char *label = 0, *pattern = 0, *rest = 0;
    char *s = str, sign = 0;

    do {
	if(*s != '"') {		/* the label */
	    if(*s == '-' || *s == '+') sign = *s++;
	    s = chop_arg(label = s, ' ');  rest = chop_arg(label, '=');
	    if(!s) break;  if(rest) { *(s - 1) = ' '; break; }
	    s = skipspace(s);
	}
	if(*s == '"') {		/* the pattern */
	    if(!(s = chop_arg(pattern = s + 1, '"'))) {
		printf("#action: pattern not \" terminated\n"); return;
	    }
	    unescape(pattern);
	    s = skipspace(s);
	}
	if(!*s) break;
	if(*s != '=') {		/* the rest */
	    printf("#action: missing a '=' after label or pattern\n"); return;
	}
	rest = s + 1;
	/* if(!brace_balance(rest)) return; */
    } while(0);

    /* at this point, either label or pattern or both are set */
    np = label ? lookup_action(label) : lookup_action_pattern(pattern);
    /*
     * FEATURE: we do not check if two actions have the same pattern.
     * Should we warn then? Deactivate the second one? hmm.
     */
    if(*np) {
	if(!rest) {
	    if(sign) {
		if(((*np)->active < 0) != (sign == '-'))
		    (*np)->active = -1 - (*np)->active;
		if(verbose)
		    printf("#action %s \"%s\" is now o%s\n", (*np)->label,
			    (*np)->pattern, (sign != '-') ? "n" : "ff");
	    } else {
		rest = insertbuf;
#if defined(_POSIX_SOURCE)
		rest += sprintf(rest, "#action %s%s \"",
			((*np)->active < 0) ? "-" : "", (*np)->label);
#else
		sprintf(rest, "#action %s%s \"",
			((*np)->active < 0) ? "-" : "", (*np)->label);
		rest += strlen(rest);
#endif
		rest += escape_specials((*np)->pattern, rest, "\\\";");
		sprintf(rest, "\"=%s", (*np)->commands);
	    }
	} else if(!*rest) {
	    if(verbose)
		printf("#action %s \"%s\" deleted\n",
			(*np)->label, (*np)->pattern);
	    delete_actionnode(np);
	    new_settings = 1;
	} else {
	    if(sign)
		if(((*np)->active < 0) != (sign != '-'))
		    sign = 0;
		else
		    (*np)->active = -1 - (*np)->active;
	    if(pattern && strcmp((*np)->pattern, pattern)) {
		str_assign(&(*np)->pattern, pattern);
		study_action_pattern(*np);
		sign = 1;
	    }
	    if(strcmp((*np)->commands, rest)) {
		str_assign(&(*np)->commands, rest);
		sign = 1;
	    }
	    if(sign) {
		if(verbose) printf("#changed action: %s%s \"%s\"=%s\n",
			((*np)->active < 0) ? "-" : "", (*np)->label,
			(*np)->pattern, rest);
		new_settings = 1;
	    }
	}
    } else if(!pattern) {
	printf("#no action labelled '%s'\n", label);
    } else if(!rest || !*rest) {
	printf("#no action with pattern \"%s\"\n", pattern);
    } else {
	static int last;
	static char lbuf[16];
	if(!label) {
	    label = lbuf;
	    do sprintf(label, "%d", ++last); while(*lookup_action(label));
	    /* reset the counter if next value will be a power of 10? */
	}
	np = add_actionnode(pattern, rest, label, sign != '-' ? 0 : -1);
	study_action_pattern(*np);
	if(verbose)
	    printf("#new action: %s \"%s\"=%s\n", label, pattern, rest);
	new_settings = 1;
    }
}

static void cmd_alias(arg)
char *arg;
{
    if(!*arg)
	show_aliases();
    else
	parse_alias(arg);
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p = aliases;
    if(p) {
	sprintf(insertbuf, "#The following alias%s defined:",
		!p->next ? " is" : "es are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#alias %s=%s", p->name, p->subst);
    } else
	printf("#No aliases are defined.\n");
}

/*
 * parse the #alias command
 */
void parse_alias(str)
char *str;
{
    aliasnode **np;
    char *rest = chop_arg(str, '=');
    if(chop_arg(str, ' ')) {
	printf("#missing a '=' after alias label.\n");
	return;
    }
    np = lookup_alias(str);
    if(*np) {
	if(!rest)
	    sprintf(insertbuf, "#alias %s=%s", str, (*np)->subst);
	else if(!*rest) {
	    if(verbose)
		printf("#deleting alias: %s=%s\n", str, (*np)->subst);
	    delete_aliasnode(np);
	    new_settings = 1;
	} else if(strcmp(rest, (*np)->subst)) {
	    /* if(!brace_balance(rest)) return; */
	    str_assign(&(*np)->subst, rest);
	    if(verbose)
		printf("#changed alias: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#unknown alias: %s\n", str);
    } else if(!*str || !strcmp(str, "#alias")) {
	printf("#won't alias '%s'\n", str);
    } else {
	/* if(!brace_balance(rest)) return; */
	add_aliasnode(str, rest);
	if(verbose)
	    printf("#new alias: %s=%s\n", str, rest);
	new_settings = 1;
    }
}

static void cmd_bind(arg)
char *arg;
{
    if(!*arg)
	show_keybinds();
    else
	parse_bind(arg);
}

/*
 * list keyboard bindings
 */
static void show_keybinds()
{
    keynode *p;
    int count = 0;
    for(p = keydefs; p; p = p->next)
	if(p->call_data) {
	    if(!count++)
		to_pager("#User-defined keys:");
	    sprintf(insertbuf, "#bind %s=%s", p->name, p->call_data);
	    to_pager(insertbuf);
	}
    if(!count)
	printf("#No user-defined key bindings.\n");
}

/*
 * parse the #bind command
 */
void parse_bind(str)
char *str;
{
    keynode **np;
    char *text;
    char *rest = chop_arg(str, '=');
    char *kseq = chop_arg(str, ' ');

    if(kseq)
	if(!*(kseq = skipspace(kseq))) kseq = 0;
	else unescape_seq(kseq);
    np = lookup_key(str);
    if(*np) {
	if(!(text = (*np)->call_data) &&
		!(text = keyfunct_name((*np)->funct))) return;
	if(kseq && strcmp(kseq, (*np)->sequence))
	    printf("#key '%s' already bound; will not rebind.\n", str);
	else if(!rest)
	    sprintf(insertbuf, "#bind %s=%s", str, text);
	else if(!*rest) {
	    if(verbose)
		printf("#key binding deleted: %s=%s\n", str, text);
	    delete_keynode(np);
	    new_settings = 1;
	} else if(strcmp(rest, text)) {
	    /* if(!brace_balance(rest)) return; */
	    (*np)->funct = run_command;
	    keybind_modify(*np, rest);
	    if(verbose)
		printf("#redefined key: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#no such key: '%s'\n", str);
    } else if(kseq ? !key_collision(kseq, 1) :
	    !key_collision(kseq = read_seq(), 0)) {
	/* NOTE: we should not allow read_seq in every situation */
	/* if(!brace_balance(rest)) return; */
	np = add_keynode(str, kseq, run_command, (char *)0);
	keybind_modify(*np, rest);
	if(verbose)
	    printf("#new key binding: %s=%s\n", str, rest);
	new_settings = 1;
    }
}

/*
 * given a pointer to edit function, return the name of the function:
 * e.g. if(f == prev_char) return "&prev-char"
 */
static char *keyfunct_name(f)
void (*f)();
{
    const struct defbind_node *p = defbinds;
    while(p->funct && p->funct != f) p++;
    if(!p->name)
	printf("#key '%s' is predefined and cannot be modified.\n", p->label);
    return p->name;
}

/*
 * assign keybind's call_data and change funct if necessary.
 * if keybind is found to be a default one, null call_data.
 */
static void keybind_modify(np, text)
keynode *np;
char *text;
{
    const struct defbind_node *p;
    if(*text == '&')
	for(p = defbinds; p->label[0]; p++)
	    if(p->name && !strcmp(p->name, text)) {
		np->funct = p->funct;
		if(p->seq && !strcmp(np->sequence, p->seq) &&
			!strcmp(np->name, p->label)) {
		    text = 0;  break;
		}
	    }
    str_assign(&np->call_data, text);
}

/*
 * determine if a keysequence collides with that of another key.
 */
static int key_collision(keyseq, shadow)
char *keyseq;
int shadow;
{
    keynode **np;
    if(!*keyseq || isprint(*keyseq)) {
	printf("#that is not a redefineable key.\n");
	return 1;
    }
    for(np = &keydefs; *np; np = &(*np)->next) {
	register char *a = (*np)->sequence;
	register char *b = keyseq;
	do if(!*a || !*b) {
	    if(*a || *b) {
		printf("#colliding keysequence: key '%s'\n", (*np)->name);
	    } else if((a = (*np)->call_data) ||
		    ((a = keyfunct_name((*np)->funct)) && !shadow)) {
		printf("#key already bound as: %s=%s\n", (*np)->name, a);
	    } else if(a) {
		delete_keynode(np);  return 0;
	    }
	    return 1;
	} while(*a++ == *b++);
    }
    return 0;
}

/*
 * prompt the user to press a key, read an escape sequence from the
 * keyboard, print it. return static pointer to sequence.
 */
static char *read_seq()
{
    static unsigned char kbuf[CAPLEN];
    int i, k;
    printf("#please press the key you want to define:");  fflush(stdout);
    /*
     * NOTE: to allow the user to set "meta keys" when the terminal has
     * no meta mode, there's no timeout after the leading escape.
     * NOTE: NUL's in key sequences are converted to \200's.
     */
    for(i = k = 0; i < CAPLEN - 1 && get_one_char(kbuf + i, k); i++) {
#if !defined(VT100) && !defined(XASCII)
	if(metakey && (kbuf[i] & 0200)) { /* metakey has the high bit set */
	    kbuf[i] &= 0177;
	    /* precede the sequence with an escape */
	    if(!i) { kbuf[1] = kbuf[0]; kbuf[0] = '\033'; i++; }
	}
#endif
	if(i || kbuf[0] != '\033') k = 1;
	/* if(!kbuf[k]) kbuf[k] = '\200'; */
    }
    kbuf[i] = 0;
    /* use escape_seq(), and print the result instead? */ 
    for(k = 0; k < i; k++)
	if(kbuf[k] == '\033') {
	    printf(" esc");
	} else if(kbuf[k] == ' ') {
	    printf(" space");
	} else if(kbuf[k] < ' ') {
	    printf(" ^%c", kbuf[k] | '@');
	    if(!kbuf[k]) kbuf[k] = '\200';
	} else if(!isprint(kbuf[k])) {
	    printf(" \\%03o", (int)kbuf[k]);
	} else
	    printf(" %c", kbuf[k]);
    newline();
    return (char *)kbuf;
}

/*
 * read a single character from stdin.
 * timeout == 0 means wait indefinitely (no timeout).
 * return 1 if a char was read, or 0 if not.
 */
static int get_one_char(cp, timeout)
char *cp;
int timeout;
{
    int err;
    fd_set fds;
    struct timeval timeoutbuf, *to;
    to = timeout ? &timeoutbuf : 0;
    do {
	FD_ZERO(&fds);	/* this is probably redundant */
	FD_SET(0, &fds);
	timeoutbuf.tv_sec = KBD_TIMEOUT / 1000000;
	timeoutbuf.tv_usec = KBD_TIMEOUT % 1000000;
	err = select(1, &fds, (fd_set *)0, (fd_set *)0, to);
    } while(err < 0 && errno == EINTR);
    if(err < 0) syserr("select");
    if(err && read(0, cp, 1) != 1)
	syserr("read from stdin");
    return err;
}

static void cmd_cancel(arg)
char *arg;
{
    editsess *sp = edit_sess;
    int num;
    if(!sp) {
	printf("#no active sessions to cancel.\n");
    } else if(*arg) {
	if((num = atoi(arg)))
	    for( ; sp; sp = sp->next)
		if(num == sp->num) {
		    cancel_edit(sp);
		    return;
		}
	printf("#no such session - use the session number\n");
    } else if(sp->next) {
	printf("#several sessions active, use #cancel <number>\n");
    } else
	cancel_edit(sp);
}

static void cmd_capture(arg)
char *arg;
{
    if(capfile) {
	if(*arg)
	    printf("#capture already active.\n");
	else {
	    fclose(capfile);  capfile = 0;
	    printf("#end of capture to file.\n");
	}
    } else {
	char *mode = "w";
	if(*arg == '>') {
	    arg = skipspace(arg + 1);
	    mode = "a";
	}
	if(!*arg)
	    printf("#capture to what file?\n");
	else if(!(capfile = fopen(arg, mode)))
	    printf("#error writing file %s\n", arg);
	else
	    printf("#capture to %s begins.\n", arg);
    }
}

static void cmd_echo(arg)
char *arg;
{
    char *pat = "%s\n";
    if(arg[0] == '-' && arg[1] == 'n') {		/* #echo -n ... */
	if(!arg[2]) return;
	if(arg[2] == ' ') {
	    pat = "%s";
	    arg += 3;
	}
    }
    unescape_seq(arg);
    printf(pat, arg);
}

/*ARGSUSED*/
static void cmd_edit(arg)
char *arg;
{
    editsess *s = edit_sess;
    if(s) {
	do if(s->state) {
	    sprintf(insertbuf, "#%s session %d '%s'",
		    s->mode == 'E' ? "editing" :
		    s->mode == 'V' ? "viewing" : "cancelled",
		    s->num, s->desc);
	    to_pager(insertbuf);
	} while((s = s->next));
    } else
	printf("#No active editors.\n");
}

static void cmd_for(arg)
char *arg;
{
    /* NOTE: #for command (and others) should be interruptable */
    if(!isdigit(*arg)) {
	char *var, *cmd = chop_arg(arg, ' ');
	int v, last = -1;
	if(cmd) {
	    while((var = get_env(arg)) && (v = atoi(var)) &&
		    (v < 0 ? v = -v : v) < (unsigned)last) {
		last = v;
		parse(cmd, 0, (struct param *)0);
	    }
	    return;
	}
    } else {
	char buf[24], *s;
	int i, to, step;
	if((i = strtol(arg, &s, 10), *s == ' ') &&
		(to = strtol(s + 1, &s, 10), !*s || *s == ' ')) {
	    if(!*s || !(step = strtol(++s, &arg, 10)) || *arg != ' ')
		step = 1;
	    else
		s = arg + 1;
	    while((step > 0) ? (i <= to) : (i >= to)) {
		sprintf(buf, "%d", i);
		set_env("I", buf);
		parse(s, 0, (struct param *)0);
		i += step;
	    }
	    set_env("I", (char *)0);
	    return;
	}
    }
    printf("#for: bad syntax\n");
}

static void cmd_hilite(arg)
char *arg;
{
    parse_hilite(arg);
}

void parse_hilite(str)
char *str;
{
    int attr = parse_attributes(str);
    if(attr >= 0) {
	/*
	 * NOTE: #hilite with !verbose (e.g. an action-bound command)
	 * could have a separate function.
	 */
	attr_string(edattr = attr, edattrbeg, edattrend);
	printf("#input highlighting is now %so%s%s.\n",
		edattrbeg, (edattr == NOATTRCODE) ? "ff" : "n", edattrend);
	new_settings = 1;
    }
}

/*ARGSUSED*/
static void cmd_help(arg)
char *arg;
{
    const cmdstruct *p;
    char *buf = insertbuf;
#ifdef HELPFILE
    if(*arg) {
	FILE *f = fopen(HELPFILE, "r");
	if(!f) {
	    printf("#can't open help file %s\n", HELPFILE);
	    return;
	}
	if(*arg == '#') arg++;
	while(fgets(buf, LINELEN, f)) if(buf[0] == '@') {
	    chop_str(buf + 1, '\n');
	    if(!buf[1] || !strcmp(arg, buf + 1)) {
		while(fgets(buf, LINELEN, f) && buf[0] != '@')
		    to_pager(buf);
		break;
	    }
	}
	fclose(f);
	return;
    }
#endif
    to_pager("# commands available:");
    for(p = commands; p->name; p++) {
	sprintf(buf, "#%s %s", p->name, p->help);
	to_pager(buf);
    }
}

static void cmd_identify(arg)
char *arg;
{
    char *p = *arg ? chop_arg(arg, ' ') : (arg = 0);
    str_assign(&edit_start, arg);
    str_assign(&edit_end, p);
    mp_active = 1;
    /* MUME doesn't prompt ident string */
    send_to_host(MCIDENT, (int)strlen(MCIDENT));
}

static void cmd_init(arg)
char *arg;
{
    if(!*arg) {
	if(!initstr)
	    printf("#init string not defined.\n");
	else
	    sprintf(insertbuf, "#init %s", initstr);
    } else if(arg[1] || arg[0] != '-') {
	if(!initstr || strcmp(arg, initstr)) {
	    /* if(!brace_balance(arg)) return; */
	    printf("#init string changed.\n");
	    str_assign(&initstr, arg);
	    new_settings = 1;
	}
    } else if(initstr) {
	printf("#init string cleared.\n");
	str_assign(&initstr, (char *)0);
	new_settings = 1;
    }
}

static void cmd_map(arg)
char *arg;
{
    if(arg[0] == '+' && arg[1] && strchr("nsewud", arg[1]))
	add_dir_to_map(arg[1]);
    else if(*arg && !*skipnum(arg))
	remove_steps(atoi(arg)); /* remove some steps without walking */
    else
	show_automap();
}

static void cmd_mark(arg)
char *arg;
{
    if(!*arg)
	show_marks();
    else
	parse_marker(arg);
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p = markers;
    if(p) {
	sprintf(insertbuf, "#The following marker%s defined:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#mark %s=%s",
		    p->pattern, attr_name(p->attrcode));
    } else
	printf("#No markers are defined.\n");
}

/*
 * parse arguments to the #mark command
 */
void parse_marker(str)
char *str;
{
    int attrcode;
    marknode **np;
    char *rest = chop_arg(str, '=');
    unescape(str);
    np = lookup_marker(str);
    if(*np) {
	if(!rest) {
	    strcpy(insertbuf, "#mark ");
	    rest = insertbuf + 6;
	    if(*str == ' ') *rest++ = '\\';
	    rest += escape_specials(str, rest, "\\=;");
	    sprintf(rest, "=%s", attr_name((*np)->attrcode));
	} else if(!*rest) {
	    if(verbose)
		printf("#mark deleted: %s=%s\n",
			(*np)->pattern, attr_name((*np)->attrcode));
	    delete_marknode(np);
	    new_settings = 1;
	} else if((attrcode = parse_attributes(rest)) >= 0 &&
		(*np)->attrcode != attrcode) {
	    (*np)->attrcode = attrcode;
	    if(verbose)
		printf("#changed mark: %s=%s\n",
			str, attr_name(attrcode));
	    new_settings = 1;
	}
    } else if(!rest || !*rest || !*str) {
	printf("#unknown marker: '%s'\n", str);
    } else if((attrcode = parse_attributes(rest)) >= 0) {
	np = add_marknode(str, attrcode);
	if(verbose)
	    printf("#new mark: %s=%s\n", str, attr_name(attrcode));
	new_settings = 1;
    }
}

/*
 * return a static pointer to name of given attribute code
 */
const char *attr_name(attrcode)
int attrcode;
{
    static char name[32];	/* should be enough */
    attr_token fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);

    if(bg == ATTR_NONE)
	return attr_names[fg];
    if(fg < ATTR_COLOURS)
	sprintf(name, "%s on %s", attr_names[fg], attr_names[bg]);
    else
	sprintf(name, "on %s", attr_names[bg]);
    return name;
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
static int parse_attributes(line)
char *line;
{
    attr_token t, bg;
    for(t = bg = ATTR_NONE; ; ) {
	if(*(line = skipspace(line))) {
	    if((t = get_attr_token(&line)) == ATTR_ERROR) break;
	    if(t == ATTR_FILL || (t < ATTR_COLOURS && line)) {
		if(t != ATTR_FILL && get_attr_token(&line) != ATTR_FILL) break;
		if(!line) break;
		if((bg = get_attr_token(&line)) >= ATTR_COLOURS) break;
	    }
	    if(line) break;
	}
	return ATTRCODE(t, bg);
    }
    /* NOTE: should we print these lines using to_pager() ? */
    printf("#invalid attribute syntax\n");
    printf("Valid arguments are:\tbold, underline, inverse\n");
    printf("\tor:\t\t[foreground] [on background]\nColours:");
    for(t = 0; t < ATTR_COLOURS; t++)
	printf(" %s%c", attr_names[t], (t + 1 == ATTR_COLOURS) ? '\n' : ',');
    return -1;
}

static void cmd_option(arg)
char *arg;
{
    if(!*arg)
	print_options(insertbuf);
    else
	parse_options(arg);
}

void parse_options(str)
char *str;
{
    char new[(opt_token)OPT_ERROR];
    opt_token k;
    for(k = 0; k < OPT_ERROR; k++)
	new[k] = 0;
    for( ; ; ) {
	int n;
	char *s = str++;
	if(*s != '-' && *s != '+') break;
	k = (opt_token)get_token(&str, option_names);
	if(k == OPT_ERROR) break;
	new[k] = *s;
	if(str) continue;
	for(n = 0, k = 0; k < OPT_ERROR; k++) {
	    char val = new[k];
	    if(!val || (val = val != '-') == OPTION(k)) continue;
	    OPTION(k) = val;
	    if(verbose)
		printf((++n == 1 ? "#options changed: %c%s" : ", %c%s"),
			new[k], option_names[k]);
	}
	if(n) { 
	    newline();
	    new_settings = 1;
	}
	return;
    }
    printf("#syntax: #option [+option|-option] [...]\n");
    printf("Valid options:");
    for(k = 0; k < OPT_ERROR; k++) {
	if(!(k % 4)) printf("\n    ");
	printf("%s%c", option_names[k], (k + 1 == OPT_ERROR) ? '\n' : ' ');
    }
}

void print_options(buf)
char *buf;
{
    opt_token k;
    char *s;
    strcpy(buf, "#option");  s = buf + 7;
    for(k = 0; k < OPT_ERROR; k++) {
	*s++ = ' ';
	*s++ = OPTION(k) ? '+' : '-';
	s += strlen(strcpy(s, option_names[k]));
    }
}

static void cmd_prefix(arg)
char *arg;
{
    if(*arg)
	str_assign(&prefixstr, arg);
    else if(!prefixstr)
	printf("#prefix not set.\n");
    else {
	if(verbose) printf("#prefix cleared.\n");
	str_assign(&prefixstr, (char *)0);
    }
}

static void cmd_quote(arg)
char *arg;
{
    /*
     * TODO: "#quote <file". like verbatim mode, but after each enter
     * (or down key on the bottom of the history) a new input line is
     * read from the file.
     */
    toggle((opt_token)OPT_VERBATIM, arg);
    printf("#%s mode.\n", OPTION(OPT_VERBATIM) ? "verbatim" : "normal");
}

static void cmd_read(arg)
char *arg;
{
    FILE *fp;
    char buf[LINELEN];
    if(*arg) {
	/* NOTE: some systems may need a fflush() before popen()? */
	fp = (arg[0] == '|') ? popen(arg + 1, "r") : fopen(arg, "r");
	if(fp) {
	    while(fgets(buf, LINELEN, fp)) {
		send_to_host(buf, (int)strlen(buf));
		expect_prompt++;
	    }
	    if(arg[0] == '|') pclose(fp); else fclose(fp);
	} else
	    printf("#error opening %s\n", arg);
    } else
	printf("#syntax: #read filename or #read |command\n");
}

static void cmd_record(arg)
char *arg;
{
    if(*arg) {
	if(recording_name) {
	    printf("#already recording to alias '%s'\n", recording_name);
	    return;
	}
	chop_arg(arg, ' ');	/* strip any accidental trailing blanks */
	str_assign(&recording_name, arg);
	if(verbose)
	    printf("#recording into alias '%s'\n", arg);
    } else if(recording_name) {
	if(verbose)
	    printf("#recording of alias '%s' stopped.\n", recording_name);
	str_assign(&recording_name, (char *)0);
	new_settings = 1;
    } else
	printf("#not recording any alias\n");
}

static void cmd_retrace(arg)
char *arg;
{
    retrace_steps(atoi(arg));
}

#ifndef OMIT_SCHEDULE
static void cmd_schedule(arg)
char *arg;
{
    if(!*arg)
	show_schedule();
    else
	parse_schedule(arg);
}

/*
 * show scheduled commands
 */
static void show_schedule()
{
    schedulenode *p = schedule;
    if(p) {
	long delay = 0;
	sprintf(insertbuf, "#The following command%s scheduled:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next) {
	    int t = (delay += p->delta) / SCH_HZ;
	    sprintf(insertbuf, "#schedule %-8s %2d:%.2d:%.2d=%s", p->label,
		    t / 3600, t / 60 % 60, t % 60, p->commands);
	}
    } else
	printf("#No commands are scheduled.\n");
}

/*
 * parse the #schedule command
 */
static void parse_schedule(str)
char *str;
{
    long delay = 0;
    schedulenode **np;
    char *rest = chop_arg(str, '=');
    char *k = chop_arg(str, ' ');
    np = lookup_schedule(str);
    if(!rest || !*rest) {
	if(*np) {
	    if(rest) {
		if(verbose)
		    printf("#scheduled command deleted: %s=%s\n",
			    str, (*np)->commands);
		delete_schedulenode(np);
	    } else if(!k) {
		sprintf(insertbuf, "#schedule %s=%s", str, (*np)->commands);
	    } else if((delay = parse_delay(k))) {
		np = reschedule(np, delay);
		if(verbose)
		    printf("#command rescheduled: %s %s=%s\n",
			    str, k, (*np)->commands);
	    }
	} else
	    printf("#no scheduled command with label '%s'\n", str);
    /* } else if(!brace_balance(rest)) { return; */
    } else if(!*np) {
	if((delay = parse_delay(k ? k : str))) {
	    static int last;
	    static char lbuf[16];
	    if(!k) {
		str = lbuf;
		do sprintf(str, "_%d", ++last); while(*lookup_schedule(str));
		/* reset the counter if next value is a power of 10? */
	    }
	    np = add_schedulenode(str, rest, delay);
	    if(verbose)
		printf("#new command scheduled: %s=%s\n", str, rest);
	}
    } else if(!k || (delay = parse_delay(k))) {
	if(strcmp((*np)->commands, rest))
	    str_assign(&(*np)->commands, rest);
	else if(!delay)
	    return;
	if(delay) reschedule(np, delay);
	if(verbose)
	    printf("#scheduled command changed: %s=%s\n", str, rest);
    }
}

/*
 * parse the command delay specification: [[hh:]mm:]ss
 * return the delay in units of 1/SCH_HZ of a second; 0 signifies error.
 */
static long parse_delay(str)
char *str;
{
    char *s = str;
    long r = 0;
    int sign = 0;
    if(*s == '+') s++, sign = 1;
    for( ; ; ) {
	if(*s) {
	    if(isdigit(*s)) r += strtol(s, &str, 10), s = str;
	    else if(*s != ':') break;
	    else s++, r *= 60;
	} else if(r) {
	    r *= SCH_HZ;
	    return sign ? -r : r;
	} else
	    break;
    }
    printf("#schedule: invalid delay specification\n");
    return 0;
}
#endif

static void cmd_set(arg)
char *arg;
{
    /*
     * NOTE: perhaps we should keep certain buffers in our environment,
     * allowing user to directly set them, e.g. "#set FOOTNOTE=HELLO!"
     */
    char *s;
    if(!*arg)
	show_variables();
    else if((s = strchr(arg, '=')) && (*s = 0, !*++s))
	set_env(arg, (char *)0);
    else
	set_env(arg, s ? s : "");
}

static void show_variables()
{
    int i;
    for(i = 0; environ[i]; i++)
	to_pager(environ[i]);
}

#ifndef NOSHELL
static void cmd_shell(arg)
char *arg;
{
    if(*arg) {
	reset_terminal();
	execute(arg);
	set_terminal();
	lastline();
    }
}
#endif

static void cmd_speedwalk(arg)
char *arg;
{
    if(toggle((opt_token)OPT_SPEEDWALK, arg) || !parse_walk(arg))
	printf("#speedwalk is now o%s.\n", OPTION(OPT_SPEEDWALK) ? "n" : "ff");
}

#ifdef TELNETBUG
static void cmd_stdcolour(arg)
char *arg;
{
    int attrcode = parse_attributes(arg);
    if(attrcode >= 0) {
	attr_token bg = BACKGROUND(attrcode), fg = FOREGROUND(attrcode);
	if(fg < ATTR_COLOURS && bg < ATTR_COLOURS) {
	    sprintf(modenorm, "\033[;3%d;4%dm", fg, bg);
	    printf("%s#standard colour set.\n", modenorm);
	    if(edattrend[0]) strcpy(edattrend, modenorm);
	} else
	    printf("#please specify both foreground and background colours\n");
    }
}
#endif

/*ARGSUSED*/
static void cmd_ver(arg)
char *arg;
{
    printver();
}

static void cmd_wrap(arg)
char *arg;
{
    toggle((opt_token)OPT_WORDWRAP, arg);
    printf("#word wrap %sabled.\n", OPTION(OPT_WORDWRAP) ? "en" : "dis");
}

#ifndef OMIT_SESSIONS
static void cmd_connect(arg)
char *arg;
{
    if(!*arg)
	show_connections();
    else {
	char *host = chop_arg(arg, '=');
	char *init = chop_arg(arg, ' ');
	if(find_sess(arg, 0))
	    printf("#session with that name already exists.\n");
	else if(!host)
	    connect_sess(arg, init, (char *)0, 0);
	else {
	    char *port = chop_arg((host = skipspace(host)), ' ');
	    connect_sess(arg, init, host, port ? atoi(port) : 0);
	}
    }
}

static void cmd_snoop(arg)
char *arg;
{
    if(!*arg)
	printf("#snoop: no session name\n");
    else
	snoop_sess(arg);
}

static void cmd_zap(arg)
char *arg;
{
    int sfd;
    if(!*arg)
	printf("#zap: no session name\n");
    else if((sfd = find_sess(arg, 1)))
	disconnect_sess(sfd);
}
#endif

/*
 * if arg is either "on" or "off", set the appropriate option, and return 1.
 * otherwise return 0.
 */
static int toggle(k, arg)
opt_token k;
char *arg;
{
    if(!strcmp(arg, "on"))
	OPTION(k) = 1;
    else if(!strcmp(arg, "off"))
	OPTION(k) = 0;
    else
	return 0;
    return 1;
}
