/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"

int strncasecmp();

static void prev_line(), next_line();
static void prev_char(), next_char();
static void prev_word(), next_word();
static void del_char_right(), del_char_left();
static void del_word_right(), del_word_left();
static void end_of_line(), beginning_of_line();
static void kill_to_eol(), redraw_input_line();
static void transpose_chars(), transpose_words();
static void enter_into_history(), enter_line();
static void complete_word(), complete_line();
static void matching_bracket();
static void downcase_word(), upcase_word();
static void capitalize_word(), set_word_case();

static keynode *search_key();
static void flashparen();
static void put_in_history();
static void swap_history();
static void to_edline();
static int skip_word();
static int skip_nonword();

/* circular word completion list */
static void demote_word();
static void delete_dup_words();

#define WORD_UNIQUE	1		/* unique among preceding words */
#define WORD_RETAIN	2		/* permanent entry (a #command) */
#define WORD_CASE	4		/* use case sensitive matching	*/

static unsigned int wordindex;
static char *words[WORDSKEPT];
static char wordflags[WORDSKEPT];

#if (WORDSKEPT - 1) >> 8
static unsigned int wordprev[WORDSKEPT];	/* older word */
static unsigned int wordnext[WORDSKEPT];	/* newer word */
#else
static unsigned char wordprev[WORDSKEPT];
static unsigned char wordnext[WORDSKEPT];
#endif

/* history buffer */
static char *hist[MAXHIST];	/* saved history lines */
static unsigned int curline;	/* current history line */
static unsigned int pickline;	/* line to pick history from */

#if (MAXHIST - 1) & MAXHIST
#define HISTINDEX(i) (((i) + MAXHIST) % MAXHIST)
#else
#define HISTINDEX(i) ((i) & (MAXHIST - 1))
#endif

static void (*last_edit_func)();	/* last editor func that was used */

/*
 * TODO:
 * C-o  operate-and-get-next
 * C-r  reverse-search-history
 * M-<  beginning-of-history
 * M->  end-of-history
 * M-?  possible-completions 
 */

const struct defbind_node defbinds[] = {
  { "enter",	"\n",		0,			enter_line },
  { "return",	"\r",		0,			enter_line },
  { "bs",	"\b",		"&del-char-left",	del_char_left },
  { "del",	"\177",		"&del-char-left",	del_char_left },
  { "tab",	"\t",		"&complete-word",	complete_word },
  { "left",	key_left,	"&prev-char",		prev_char },
  { "right",	key_right,	"&next-char",		next_char },
  { "up",	key_up,		"&prev-line",		prev_line },
  { "down",	key_down,	"&next-line",		next_line },
  { "C-a",	"\001",		"&beginning-of-line",	beginning_of_line },
  { "C-b",	"\002",		"&prev-char",		prev_char },
  { "C-c",	"\003",		0,			intrsig },
  { "C-d",	"\004",		"&del-char-right",	del_char_right },
  { "C-e",	"\005",		"&end-of-line",		end_of_line },
  { "C-f",	"\006",		"&next-char",		next_char },
  { "C-g",	"\007",		"&matching-bracket",	matching_bracket },
  { "C-k",	"\013",		"&kill-to-eol",		kill_to_eol },
  { "C-l",	"\014",		"&redraw-input-line",	redraw_input_line },
  { "C-n",	"\016",		"&next-line",		next_line },
  { "C-p",	"\020",		"&prev-line",		prev_line },
  { "C-t",	"\024",		"&transpose-chars",	transpose_chars },
  { "C-y",	"\031",		"&enter-into-history",	enter_into_history },
  { "C-z",	"\032",		0,			suspend_cancan },
  { "M-del",	"\033\177",	"&del-word-left",	del_word_left },
  { "M-tab",	"\033\t",	"&complete-line",	complete_line },
  { "M-b",	"\033b",	"&prev-word",		prev_word },
  { "M-c",	"\033c",	"&capitalize-word",	capitalize_word },
  { "M-d",	"\033d",	"&del-word-right",	del_word_right },
  { "M-f",	"\033f",	"&next-word",		next_word },
  { "M-l",	"\033l",	"&downcase-word",	downcase_word },
  { "M-t",	"\033t",	"&transpose-words",	transpose_words },
  { "M-u",	"\033u",	"&upcase-word",		upcase_word },
  { "",		0,		0,			0 }
};

/*
 * the line editor. given some input chars (typically a single one),
 * add them to edit line, or call appropriate edit funcs.
 */
void line_edit(s, len)
char *s; int len;
{
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static int nchars;		/* the number of them */
    int i;
    keynode *p;
    if(!(linemode & LM_NOECHO)) {	/* normal line mode */
	for(i = 0; i < len; i++) {
	    unsigned char c = s[i];
	    /*
	     * NOTE: currently, we insertchars() or search_key()
	     * with every char separately. This is ugly.
	     */
	    if(pager_text && pager(c))
		continue;
	    if(!nchars && isprint(c)) {
		if(edlen < LINELEN - 1) {
		    insertchars(s + i, 1);
		    if(OPTION(OPT_FLASHPAREN) && isrparen(c))
			flashparen();
		    last_edit_func = 0;
		}
		continue;
	    }
#if !defined(VT100) && !defined(XASCII)
	    if(metakey && (c & 0200)) {	/* metakey set the high bit */
		c &= 0177;
		if(!nchars)
		    typed[0] = '\033', nchars = 1;
	    }
#endif
	    typed[nchars++] = c ? c : '\200';
	    p = search_key(typed, nchars);
	    if(p) {
		if(p->sequence[nchars])	/* partial match */
		    continue;
		(*p->funct)(p->call_data);
		last_edit_func = p->funct;
	    }
	    nchars = 0;
	}
    } else {
	/*
	 * sending password (line mode, no echo)
	 * FEATURE: we still recognize C-c, but only at first pos.
	 */
	for(i = 0; i < len; i++) {
	    edbuf[edlen++] = s[i];
	    if(edbuf[0] == '\003')
		intrsig();
	    if(s[i] == '\n')
		newline();
	    else if(edlen < LINELEN)
		continue;
	    send_to_host(edbuf, edlen);
	    expect_prompt++;
	    edlen = 0;
	}
	edbuf[pos = edlen] = 0;
    }
}

/*
 * lookup a keynode with specified sequence
 */
static keynode *search_key(seq, len)
char *seq; int len;
{
    register keynode *p;
    for(p = keydefs; p; p = p->next) {
	register char c = seq[0];
	while(c != p->sequence[0] && (p = p->next));
	if(!p || len <= 1) break;
	if(!strncmp(p->sequence, seq, len)) break;
    }
    return p;
}

/*
 * Flash cursor at parenthesis matching the one inserted before current pos
 */
static void flashparen()
{
    int lev = 0, i = pos;
    char c = 0;
    while(--i >= 0)
	if(isrparen(edbuf[i])) {
	    lev++;
	    if(!c) c = edbuf[i];
	} else if(islparen(edbuf[i]) && !--lev) {
	    if(LRPARENS(edbuf[i], c)) {
		attention |= ATTN_FLASHBACK;
		flashback = pos;
		moveto(i);
	    }
	    break;
	}
}

/*
 * transpose the char under the cursor with previous one
 */
static void transpose_chars()
{
    if(edlen > 1) {
	char c;
	int k = pos;
	if(k && k-- >= edlen) k--;
	moveto(k);
	c = edbuf[k];
	overtype_char(edbuf[k + 1]);
	overtype_char(c);
    }
}

/*
 * transpose the word behind the cursor with next one
 */
static void transpose_words()
{
    int a, b, c;
    prev_word();
    a = skip_word(skip_nonword(pos));
    b = skip_nonword(a);
    c = skip_word(b);
    if(b < c) {
	char buf[LINELEN], *s;
	sprintf(buf, "%.*s%.*s%.*s",
		c - b, edbuf + b, b - a, edbuf + a, a - pos, edbuf + pos);
	s = buf; do overtype_char(*s); while(*++s);
    }
}

/*
 * set the word in lowercase, move to the end of it
 */
static void downcase_word()
{
    set_word_case(0);
}

/*
 * set the word in uppercase, move to the end of it
 */
static void upcase_word()
{
    set_word_case(LINELEN);
}

/*
 * capitalize a word, and move to the end of it
 */
static void capitalize_word()
{
    set_word_case(1);
}

/*
 * ncaps is the number of leading chars to put in uppercase.
 * rest of the word is downcased.
 */
static void set_word_case(ncaps)
int ncaps;
{
    int b, a = skip_nonword(pos);
    moveto(a);
    for(b = skip_word(a); a < b; a++) {
	int c = (unsigned char)edbuf[a];
	overtype_char(--ncaps >= 0 ? toupper(c) : tolower(c));
    }
}

/*
 * redraw the input line
 */
static void redraw_input_line()
{
    clear_all();
    col0 = 0;	/* nuke the prompt */
    redraw_input();
}

/*
 * erase everything to the end of line
 */
static void kill_to_eol()
{
    int n = edlen - pos;
    if(n) deletechars(n);
}

/*
 * move cursor to end of line
 */
static void end_of_line()
{
    moveto(edlen);
}

/*
 * move cursor to beginning of line
 */
static void beginning_of_line()
{
    moveto(0);
}

/*
 * delete a character to the right
 */
static void del_char_right()
{
    if(pos < edlen)
	deletechars(1);
}

/*
 * delete a character to the left
 */
static void del_char_left()
{
    if(pos) {
	moveto(pos - 1);
	deletechars(1);
    }
}

/*
 * move the editline into history
 */
static void enter_into_history()
{
    if(edlen) {
	put_in_history(edbuf);
	pickline = curline;
	to_edline("");
    }
}

/*
 * enter a line
 */
static void enter_line()
{
    skip_edline();
    insertbuf[0] = '\0';
    if(capfile) to_capfile(edbuf, 1);
    if(OPTION(OPT_VERBATIM) && edbuf[0] != '#' && edbuf[0] != '{') {
	/* NOTE: should we skip leading '\' in verbatim mode? */
	edbuf[edlen] = '\n';
	send_to_host(edbuf, edlen + 1);
	expect_prompt++;
	edbuf[edlen] = '\0';
    } else {
	new_settings = 0;
	verbose = 1;
	/*
	 * FEATURE: we should check autosave flag both before and
	 * after parse(), so that #option -autosave is saved too..
	 */
	parse(edbuf, 1, (struct param *)0);
	verbose = 0;
	/* save if user changed or defined something */
	if(new_settings && OPTION(OPT_AUTOSAVE))
	    save_settings();
    }
    if(edlen > 1) {
	/* don't put identical lines in history, nor short ones */
	char *p = hist[HISTINDEX(curline - 1)];
	if(!p || strcmp(edbuf, p)) put_in_history(edbuf);
    }
    pickline = curline;
    edbuf[0] = '\0';
    if(!pager_text && insertbuf[0])
	escape_command(insertbuf, edbuf);
    else if(prefixstr)
	strcpy(edbuf, prefixstr);
    pos = edlen = strlen(edbuf);
    redraw_all();
}

/*
 * move one word forward
 */
static void next_word()
{
    moveto(skip_word(skip_nonword(pos)));
}

/*
 * move one word backward
 */
static void prev_word()
{
    int i = pos;
    if(i) {
	do --i; while(i && !islabel(edbuf[i]));
	while(i && islabel(edbuf[i - 1])) i--;
	moveto(i);
    }
}

/*
 * delete word to the right
 */
static void del_word_right()
{
    int n = skip_word(skip_nonword(pos)) - pos;
    if(n) deletechars(n);
}

/*
 * delete word to the left
 */
static void del_word_left()
{
    int i = pos;
    prev_word();
    if((i -= pos) > 0)
	deletechars(i);
}

/*
 * get previous line from history list
 */
static void prev_line()
{
    int i = HISTINDEX(pickline - 1);
    if(i != curline)
	swap_history(i);
}

/*
 * get next line from history list
 */
static void next_line()
{
    if(pickline != curline) {
	int i = HISTINDEX(pickline + 1);
	swap_history(i);
    }
}

/*
 * move one char backward
 */
static void prev_char()
{
    if(pos)
	moveto(pos - 1);
}

/*
 * move one char forward
 */
static void next_char()
{
    if(pos < edlen)
	moveto(pos + 1);
}

/*
 * move to matching bracket
 */
static void matching_bracket()
{
    char *p, *brackets = "(){}[]<>";
    int i = pos;
    if((p = strchr(brackets, edbuf[i]))) {
	int nested = 0;
	int dir = (p - brackets) & 1 ? -1 : 1;
	while((i += dir) >= 0 && i < edlen)
	    if(edbuf[i] == p[0]) nested++;
	    else if(edbuf[i] == p[dir] && --nested < 0) {
		moveto(i);
		break;
	    }
    }
}

#ifdef _POSIX_SOURCE
#define WORDCMP(flags, word, str, len)					\
    ((*(flags & WORD_CASE ? strncmp : strncasecmp))			\
	((const char *)(word), (const char *)(str), (size_t)(len)))
#else
#define WORDCMP(flags, word, str, len)					\
	(flags & WORD_CASE						\
		? strncmp(word, str, len)				\
		: strncasecmp(word, str, len))
#endif

/*
 * match and complete a word referring to the word list
 */
static void complete_word()
{
    static int curr_word, comp_len, root_len;
    char *root, *p;
    int k, n;

    /* determine curr_word, root, root_len */
    if(last_edit_func == complete_word && (k = comp_len)) {
	moveto(pos - k);
	n = pos - root_len;
    } else {
#ifdef DELIM
	for(n = pos; n && !strchr(DELIM, edbuf[n - 1]); n--);
#else
	for(n = pos; n && islabel(edbuf[n - 1]); n--);
#endif
	k = 0;  curr_word = wordindex;  root_len = pos - n;
    }
    root = edbuf + n;  comp_len = 0;

    /* look up next valid completion and print it */
    while((p = words[curr_word = wordprev[curr_word]])) {
	if(WORDCMP(wordflags[curr_word], p, root, root_len)) continue;
	if(!*(p += root_len)) continue;		/* already complete */
	if((n = strlen(p)) + edlen - k >= LINELEN) continue; /* no room */
	comp_len = n;
	/*
	 * NOTE: could change this to overtype the whole word (the
	 * beginning might have different capitalization).
	 */
	while(k ? (--k, overtype_char(*p++), --n) : (insertchars(p, n), 0));
	break;
    }
    if(k) deletechars(k);
    if(p) delete_dup_words(curr_word);
}

/*
 * match backwards in history and complete entire lines
 */
static void complete_line()
{
    if(edlen) {
	unsigned int i = curline;
	for( ; ; ) {
	    i = HISTINDEX(i - 1);
	    if(i == curline || !hist[i]) break;
	    if(!strncmp(hist[i], edbuf, edlen)) {
		to_edline(hist[i]);
		break;
	    }
	}
    }
}

/*
 * swap editline with a history line
 */
static void swap_history(line)
int line;
{
    if(hist[line]) {
#if 0	/* avoid some malloc()'s */
	if(last_edit_func != prev_line && last_edit_func != next_line)
#endif
	str_assign(hist + pickline, edbuf);
	pickline = line;
	to_edline(hist[line]);
    }
}

/*
 * clear old input line, copy new one into edbuf,
 * display it, set edlen and pos.
 */
static void to_edline(str)
char *str;
{
    clear_input();
    strcpy(edbuf, str);
    edlen = pos = strlen(str);
    if(edlen) redraw_input();
}

/*
 * return the position after leading islabel() chars
 */
static int skip_word(i)
int i;
{
    while(islabel(edbuf[i])) i++;
    return i;
}

/*
 * return the position after leading word separator chars
 */
static int skip_nonword(i)
int i;
{
    while(i < edlen && !islabel(edbuf[i])) i++;
    return i;
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
static void put_in_history(str)
char *str;
{
    char *s;
    int i;
    str_assign(hist + curline, str);
    curline = HISTINDEX(curline + 1);
    /* split into words and put into completion list (unless too short) */
    do {
#ifdef DELIM
	s = str + strspn(str, DELIM);
	if(!*s) break;
	i = strcspn(s, DELIM);
#else
	for(s = str; *s && !islabel(*s); s++);
	for(i = 0; islabel(s[i]); i++);
#endif
	str = 0;
	if(s[i]) s[i] = 0, str = s + i + 1;
	if(i >= 3) put_word(s);
    } while(str);
}

/*
 * put word in word completion ring
 */
void put_word(s)
char *s;
{
    /* TODO: allow the user to add permanent entries */
    int r = wordindex;
    words[r] = str_dup(s);  wordflags[r] = 0;
    while(wordflags[r = wordnext[r]] & WORD_RETAIN);
    demote_word(r);  wordindex = r;
    str_assign(words + r, (char *)0);
}

/*
 * reinsert a node just before wordindex ie make it the last word
 */
static void demote_word(i)
int i;
{
    wordprev[wordnext[i]] = wordprev[i];
    wordnext[wordprev[i]] = wordnext[i];
    wordnext[i] = wordnext[wordprev[i] = wordindex];
    wordnext[wordindex] = wordprev[wordnext[wordindex]] = i;
}

/*
 * delete duplicate instances of a word
 */
static void delete_dup_words(i)
int i;
{
    int n, k = i;
    char *p;
    if(!(wordflags[k] & WORD_UNIQUE)) {
	wordflags[k] += WORD_UNIQUE;
	p = words[k];  n = wordprev[k];
	while(words[n]) {
	    n = wordprev[k = n];
	    if(WORDCMP(wordflags[k], words[k], p, (unsigned)-1)) continue;
	    demote_word(k);
	    free(words[k]);  words[k] = 0;
	    wordflags[i] |= wordflags[k];	/* move word flags */
	    if((wordflags[k] &= WORD_UNIQUE)) break;
	}
    }
}

/*
 * initialize the word completion ring
 */
void default_completions()
{
    char buf[20];	/* should be enough */
    const cmdstruct *p;
    int i, k;
    /* NOTE: perhaps keep the '#' command prefix in commands struct? */
    for(i = 0, p = commands; p->name; p++)
#if defined(_POSIX_SOURCE)
	if(p->func && sprintf(buf, "#%s", p->name) > 3) {
#else
	if(p->func && strlen(p->name) >= 3) {
	    sprintf(buf, "#%s", p->name);
#endif
	    if(++i >= WORDSKEPT) break;
	    words[i] = str_dup(buf);
	    wordflags[i] = WORD_UNIQUE | WORD_RETAIN | WORD_CASE;
	}
    k = 0; i = WORDSKEPT - 1;
    do {
	wordprev[i] = k;
	wordnext[k] = i;
	k = i;
    } while(--i >= 0);
}
