/*
 * parse.c
 * command parsing, argument parsing routines for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"

FILE *popen();
int pclose();
long strtol();

static void subst_param();
#ifdef XEXPR
static char *expr_tok();
#endif
static char *getval();
static char *calc();
static void record_line();
static struct param *match_action();
static int exec_action();
static void split_arg();
static char *skip_arg();

#ifdef OWN_CTYPE
const unsigned char ct_array[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    64, 1, 1, 1, 1, 1, 1, 1,33,17, 1, 1, 1, 1, 1, 1,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1, 1, 1, 1, 1, 1,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 9,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 0,
#if defined(XASCII)
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
#endif
};
#endif

/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params are passed along.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static defnode *recursed_aliases;
    char cmd[LINELEN], buf[LINELEN];
    char *s, *d;
    int i, subst;
    aliasnode *np;
    cmd_func *func;

    if(!brace_balance(cmdline)) return;

    s = skipspace(cmdline);			/* line enclosed in braces? */
    if(*s == '{' && !*skipspace((d = skip_arg(++s, '}')) + 1))
	memcpy(cmd, s, d - s), cmd[d - s] = 0, subst = 1;
    else
	strcpy(cmd, cmdline), subst = 0;

    /* PROCESS THE COMMANDS ONE BY ONE */
    s = cmd;
    do {
	cmdline = chop_arg(s, CMDSEP);		/* chop into commands */
	if(recording_name && *s != '#' && !param) record_line(s);

	buf[0] = buf[sizeof(buf) - 1] = 0;

	if((d = skip_arg(s, ' '))) *d = 0;	/* pick out the first word */

	if((np = *lookup_alias(s)) && (*lookup(s, &recursed_aliases) ||
		(recording_name && !strcmp(s, recording_name)))) np = 0;

	if(np) { if(d) *d = ' '; }
	else if(*s == '#') { if(d) d = skipspace(d + 1); }
	else { if(d) *d = ' '; d = s; }		/* restore line */

	if(d)
	    if(subst) subst_param(buf, d, param);
	    else strcpy(buf, d);

	if(param && !cmdline && !subst && recursed_aliases && param[0].len) {
	    /* append arguments */
	    int n = sizeof(buf) - (i = strlen(buf));
	    if(param[0].len < n) n = param[0].len + 1;
	    memcpy(buf + i, param[0].s, n);
	}
	if(buf[sizeof(buf) - 1]) {
	    buf[sizeof(buf) - 1] = 0;
	    printf("#warning: line truncated\n");
	}

	if(np) {				/* process an alias */
	    struct param new_param[10];
	    defnode ban_alias;
	    ban_alias.next = recursed_aliases;  ban_alias.field = s;
	    if(d) *d = 0;			/* cut out first word again */
	    split_arg(buf, new_param);
	    recursed_aliases = &ban_alias;	/* block alias recursion */
	    parse(np->subst, silent, new_param);
	    recursed_aliases = ban_alias.next;
	} else if(*s != '#') {			/* text to send		*/
	    /* NOTE: should we echo the walk sequence also if !silent? */
	    if(buf[0]) {
		if(!buf[1] || OPTION(OPT_SPEEDWALK))
		    if(parse_walk(buf)) continue;
		if(!subst) unescape(buf);
	    }
	    if(!silent) printf("[%s%s%s]\n", edattrbeg, buf, edattrend);
	    buf[i = strlen(buf)] = '\n';
	    send_to_host(buf, i + 1);
	    expect_prompt++;
	} else if(*++s) {			/* "# " is a comment	*/
	    if(isdigit(*s) && !*skipnum(s)) {
		if((i = atoi(s)))
		    do parse(buf, 1, (struct param *)0); while(--i);
		else
		    printf("#bogus repeat count\n");
	    }
#ifndef OMIT_SESSIONS
	    else if(!is_main_sess())
		printf("#command not allowed on subsidiary connection\n");
	    else if(*s == '#') {		/* a session is specified */
		int backfd = sockfd;
		if((sockfd = find_sess(s + 1, 1))) {
		    int back_expect_prompt = expect_prompt;
		    parse(buf, 1, (struct param *)0);
		    if(!is_main_sess())
			expect_prompt = back_expect_prompt;
		}
		sockfd = backfd;
	    }
#endif
	    /*
	     * NOTE: perhaps we should keep a parse_context structure, or
	     * similar, to pass on to commands, so that #add, #? could be
	     * moved to cmd.c. Or maybe let the commands do their own args
	     * expansions a la pull_arg(). Then the double unescaping with
	     * #echo command could be avoided, etc., etc.
	     */
	    else if((func = cmdfunc_by_name(s))) {
		(*func)(buf);
		/*
		 * NOTE: to ensure we know on what line we are, lastline()
		 * may have to be used. This will leave some empty lines.
		 * Perhaps we should set a flag instead, and scroll after
		 * redrawing input, to remove the uncertainty.
		 */
		if(CURLINE(edlen) < lines_1) lastline();
		/* if(line0 < lines_1) lastline(); */
	    } else if(param && !recursed_aliases && !strcmp(s, "add")) {
		if(param[0].len) {
		    sprintf(buf, "%.*s", param[0].len, param[0].s);
		    put_word(buf);
		}
	    }
#ifndef NOSHELL
	    else if(*s == '?' && !*(s + 1)) {	/* #? command	*/
		i = execute(buf);
		if(!(s = cmdline)) break;
		cmdline = chop_arg(s, CMDSEP);
		if(!i) {
		    parse(s, silent, param);
		    return;
		}
	    }
#endif
	    else
		printf("#unknown cancan command '%s'\n", s);
	}
    } while((s = cmdline));
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk(word)
char *word;
{
    static char wcmd[2] = { ' ', '\n' };
    char *s = word;
    int n = strspn(s, "nsewud0123456789");
    if(!n || s[n] || isdigit(s[n - 1]))
	return 0;
    do {
	n = 1;
	if(isdigit(*s)) {
	    n = strtol(s, &word, 10);
	    s = word;
	    if(!n) continue;
	}
	expect_prompt++;
	wcmd[0] = *s;
	do {
	    send_to_host(wcmd, 2);
	    if(is_main_sess())		/* yargh */
		add_dir_to_map(*s);
	} while(--n);
    } while(*++s);
    return 1;
}

/*
 * verify that the braces in the line are properly balanced
 */
int brace_balance(str)
char *str;
{
    char *message = "#unbalanced %c near \"%.16s\"\n";
    char *d, *s = str;
    if(skip_arg(s, '\0')) return 1;
    if((d = skip_arg(s, '}')))
	printf(message, *d, (d - str > 15) ? d - 15 : str);
    else {
	do {
	    d = skip_arg(s, '{');
	    s = skip_arg(d + 1, '}');
	} while(s++);
	printf(message, *d, d);
    }
    return 0;
}

/*
 * {} *must* be balanced in input string. input will be trashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &[expression], &(command).
 */
static void subst_param(buf, str, param)
char *buf, *str;
struct param param[10];
{
    char *src, *s = str;
    int i, room = LINELEN;

    while(*s) {
	if(*s != '&') {
	    if(*s != '{') {
		if(*s == '\\' && !*++s) break;
		if(room) --room, *buf++ = *s++;
		continue;
	    } else {
		src = s;  s = skip_arg(s + 1, '}') + 1;
		i = s - src;
	    }
	} else switch(*++s) {
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
	    {		/* {} must be balanced, no check here	*/
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		src = get_env(sbuf);
	    }
	    if(!src) continue;
	    i = strlen(src);
	    break;
	case '[':				/* &[expr]	*/
	    s = skip_arg((src = s + 1), ']');
	    if(*s) {
		static char vbuf[24];  long val;
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		if(!(src = calc(sbuf, &val)) || *src) continue;
#if defined(_POSIX_SOURCE)
		i = sprintf(src = vbuf, "%ld", val);
#else
		sprintf(src = vbuf, "%ld", val);
		i = strlen(src);
#endif
		break;
	    } else
		printf("#unbalanced [ near \"&[%.16s\"\n", src);
	    continue;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(*s) {
		FILE *fp;
		*s++ = 0;
		unescape(src);
		/* NOTE: some systems may need a fflush() before popen()? */
		if((fp = popen(src, "r"))) {
		    if((i = fread(buf, 1, room, fp)) > 0) {
			if(buf[i - 1] == '\n') i--;
			buf += i;  room -= i;
		    }
		    pclose(fp);
		}
	    } else
		printf("#unbalanced ( near \"&(%.16s\"\n", src);
	    continue;
	default:
	    if(!islabel(*s)) {
		if(room) --room, *buf++ = '&';
		continue;
	    }
	    if(isdigit(*s)) {
		i = ctoi(*s++);
		if(!param || !param[i].len) continue;
		src = param[i].s;  i = param[i].len;
		break;
	    }
	    for(src = s; islabel(*++s); );
	    i = *s;  *s = 0;
	    src = get_env(src);  *s = i;
	    if(!src) continue;
	    i = strlen(src);
	    break;
	} /* switch ends */
	if(room < i) i = room;
	memcpy(buf, src, i);
	buf += i;  room -= i;
    } /* end while */
    if(room) *buf = 0;
}

#ifdef XEXPR
typedef enum {
    T_MUL, T_DIV, T_MOD, T_ADD, T_SUB,
    T_LT, T_GT, T_AND, T_OR, T_XOR,
    T_EQ, T_NE, T_SHL, T_SHR,
    T_LE, T_GE, T_LAND, T_LOR,
    T_NOP
} oper;

#define N_PR	11		/* number of precedence levels */
static const char op_precedence[] = {
    10, 10, 10, 9, 9,
    7, 7, 5, 3, 4,
    6, 6, 8, 8,
    7, 7, 2, 1,
    0
};

static const char op_string[] = {
    '*', '/', '%', '+', '-',
    '<', '>', '&', '|', '^',
    '=', '!', 0
};

static char *expr_tok(str, op)
char *str;
oper *op;
{
    oper c = T_NOP;
    char *s = skipspace(str);
    if(*s && (str = strchr(op_string, *s))) {
	c = str - op_string; s++;
	switch(c) {
	    case T_EQ:
	    case T_NE:
		if(*s == '=') s++;
		else c = T_NOP;
		break;
	    case T_LT:
		if(*s == '<') { c = T_SHL;  s++; }
		else if(*s == '=') { c += T_LE - T_LT;  s++; }
		break;
	    case T_GT:
		if(*s == '>') { c = T_SHR;  s++; }
		else if(*s == '=') { c += T_GE - T_GT;  s++; }
		break;
	    case T_AND:
		if(*s == '&') { c += T_LAND - T_AND;  s++; }
		break;
	    case T_OR:
		if(*s == '|') { c += T_LOR - T_OR;  s++; }
		break;
	    default:
		break;
	}
    }
    *op = c;
    return s;
}
#endif	/* XEXPR */

/*
 * read a value from string; recurse into calc() if parentheses are found.
 * return string pointer properly advanced, NULL signifies failure.
 */
static char *getval(str, val)
char *str; long *val;
{
    char *s = str;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    }
    *val = strtol(s, &str, 10);
    return s != str ? str : 0;
}

/*
 * calculate the value of a simple expression. () + - * / % are recognized.
 * return advanced string pointer, or NULL in case of syntax error.
 */
static char *calc(str, val)
char *str; long *val;
{
#ifndef XEXPR
    long v1, v2;
    char op, sign = 0;
    char *s = str;
    if(*s == '-') sign = 1, s++;
    for(*val = 0; (s = getval(s, &v1)); s++) {
	while(op = *s, op == '*' || op == '/' || op == '%')
	    if(!(s = getval(s + 1, &v2))) return 0;
	    else if(op == '*') v1 *= v2;
	    else if(!v2) return 0;
	    else if(op == '/') v1 /= v2;
	    else v1 %= v2;
	*val += sign ? -v1 : v1;
	if(op == '-') sign = 1;
	else if(sign = 0, op != '+') return s;
    }
    return 0;
#else
    long v[N_PR], b;
    oper op[N_PR], o1, o2;
    char *u_s, *s = str;
    int u_ops, n = 0;
    do {
	u_s = s;
	while(*s == ' ' || *s == '+' || *s == '-' || *s == '~' || *s == '!') s++;
	u_ops = s - u_s;
	s = getval(s, v + n);
	if(!s) break;
	/* HANDLE UNARY OPERATORS */
	while(--u_ops >= 0)
	    if(u_s[u_ops] == '-') v[n] = -v[n];
	    else if(u_s[u_ops] == '~') v[n] = ~v[n];
	    else if(u_s[u_ops] == '!') v[n] = !v[n];
	s = expr_tok(s, op + n);
	/* HANDLE BINARY OPERATORS */
	while(n && op_precedence[o2 = op[n]] <= op_precedence[o1 = op[n - 1]])
	    switch(op[--n] = o2, b = v[n + 1], o1) {
		case T_EQ: v[n] = v[n] == b;  break;
		case T_NE: v[n] = v[n] != b;  break;
		case T_LE: v[n] = v[n] <= b;  break;
		case T_GE: v[n] = v[n] >= b;  break;
		case T_LT: v[n] = v[n] < b;  break;
		case T_GT: v[n] = v[n] > b;  break;
		case T_LOR: v[n] = v[n] || b;  break;
		case T_LAND: v[n] = v[n] && b;  break;
		case T_SHL: v[n] <<= b;  break;
		case T_SHR: v[n] >>= b;  break;
		case T_OR: v[n] |= b;  break;
		case T_XOR: v[n] ^= b;  break;
		case T_AND: v[n] &= b;  break;
		case T_ADD: v[n] += b;  break;
		case T_SUB: v[n] -= b;  break;
		case T_MUL: v[n] *= b;  break;
		case T_DIV: v[n] /= b;  break;
		case T_MOD: v[n] %= b;  break;
		default:  break;
	    }
    } while(op[n++] != T_NOP);
    *val = v[0];
    return s;
#endif
}

/*
 * if alias recording is active, record given line in that alias
 */
static void record_line(line)
char *line;
{
    static int i;
    aliasnode *np;
    if(!(np = *lookup_alias(recording_name))) {
	i = 0;
	add_aliasnode(recording_name, line);
    } else {
	/*
	 * FEATURE: if both alias name and the line to be recorded
	 * are very long, total alias lenght may grow too large.
	 */
	char new_line[LINELEN];
	int room = LINELEN - 10 - strlen(line);
#if defined(_POSIX_SOURCE)
	int len = sprintf(new_line, "%s", np->subst);
#else
	int len = strlen(np->subst);
	strcpy(new_line, np->subst);
#endif
	if(len >= room - LINELEN / 2) {	/* leave ample room for editing */
	    do {
#if defined(_POSIX_SOURCE)
		len = sprintf(new_line, "%s%d", recording_name, ++i);
#else
		sprintf(new_line, "%s%d", recording_name, ++i);
		len = strlen(new_line);
#endif
	    } while(*lookup_alias(new_line));
	    if(len >= room) return;
	    add_aliasnode(new_line, np->subst);
	}
	sprintf(new_line + len, ";%s", line);
	str_assign(&np->subst, new_line);
    }
}

/*
 * scan through the pattern, determine what parameters it sets
 * and what is the length of a shortest string that can match.
 * NOTE: maybe we should go on, and compile the pattern here..
 */
void study_action_pattern(p)
actionnode *p;
{
    int mask = 0x3ff, n = 0;
    char *s = p->pattern;
    if(*s == '^') s++;
    for( ; *s; s++)
	if(*s != '&') n++;
	else {
	    unsigned int prm;
	    while(*++s == '&');
	    prm = ctoi(*s);
	    if(prm >= 10) prm = 0, s--;
	    mask &= ~(1 << prm);
	}
    if(p->active < 0) n = -1 -n;
    p->active = n;
    p->mask = mask;
}

/*
 * match action containing &0..&9 and return actual params in param array
 *
 * NOTE: if pattern contains a sequence of &n-s, all but last are ignored.
 * FEATURE: &&n will match a single word and put the result in &n.
 * FEATURE: & should not match an empty word.
 * FEATURE: we should try to find the longest possible match.
 * FEATURE: can't match '&' as plain text in pattern (its always special).
 * FEATURE: no backtracking used, therefore a match with & might fail, even
 * though respective regexp would not, e.g. pat:"a & b", src:"a a a b" fails.
 */
static struct param *match_action(pat, src)
char *pat, *src;
{
    static struct param param[10];
    char *mpat = insertbuf;	/* use insertbuf for tmp buf space. hm */
    register char *s = src, *p = pat;
    int k, prm;
    char mflags = 1;	/* 1 - at pattern beginning, 2 - single word match */
    if(*p == '^') {			/* anchor match at line start */
	if(col0) return 0;
	for(p++; *p && *p != '&'; s++, p++)
	    if(*s != *p) return 0;
	mflags = 0;
    }
    prm = -1;
    while(*p) {
	if(*p == '&') do {		/* pick out parameters */
	    p++;  prm = ctoi(*p++);
	    if((unsigned)prm >= 10)	/* support single & for back compat */
		p--, prm = 0, mflags |= 2;	/* match a single word only */
	} while(*p == '&');
	if(*p) {			/* attempt match */
	    register char *mp = mpat;
	    do *mp++ = *p++; while(*p && *p != '&');
	    *mp = '\0';
	    if(prm >= 0)
		param[prm].s = s;
	    s = strstr(s, mpat);
	    if(!s) return 0;
	    if(prm >= 0)
		param[prm].len = s - param[prm].s;
	    s += mp - mpat;

	    if(mflags & 2) {		/* post-process single word match */
		char *w;
		k = param[prm].len;  w = param[prm].s;
		while(k && islabel(w[k - 1])) k--;
		if(k) {
		    if(!(mflags & 1)) return 0;
		    /* unanchored '&' start, take last word */
		    param[prm].len -= k;  param[prm].s += k;
		}
	    }
	    mflags = 0;
	    continue;
	}
	/* '&' at the end of the pattern space */
	if(mflags & 2) {	/* take first word */
	    if(mflags & 1) while(*s && !islabel(*s)) s++;
	    for(k = 0; islabel(s[k]); k++);
	} else
	    k = strlen(s);
	param[prm].len = k;  param[prm].s = s;
	break;
    } /* end while */
    return param;
}

/*
 * execute an action. return 1 if line should be printed,
 * or 0 if it should be suppressed.
 */
static int exec_action(p, param)
actionnode *p;
struct param *param;
{
    unsigned int mask = p->mask;
    struct param *pp = param;
    if(!strcmp("#gag", p->commands)) return 0;
    while((mask >>= 1)) /* clear the params we did not set */
	if(pp++, mask & 1) pp->len = 0;
    clear_prompt();
    parse(p->commands, 0, param);
    redraw_prompt();
    return 1;
}

/*
 * search for actions to trig on an input line.
 */
int search_action(line, len)
char *line;
int len;
{
    unsigned int l = len;
    actionnode *p;
    /*
     * NOTE: a negative p->active means action is disabled. Otherwise,
     * its value is the minimum length of matchable text.
     */
    for(p = actions; p; p = p->next)
	if((unsigned)p->active <= l) {
	    struct param *match = match_action(p->pattern, line);
	    if(!match) continue;
	    /* set &0 to whole line if it wasn't used in pattern */
	    if(p->mask & 1) { match[0].len = l; match[0].s = line; }
	    return exec_action(p, match);
	}
    return 1;
}

/*
 * escape special chars with a backslash
 * return the length of resulting string
 */
int escape_specials(str, buf, spec)
char *str, *buf, *spec;
{
    int k;
    char *s, *d;
    /* NOTE: we don't check for overflow here, it shouldn't happen */
    for(s = str, d = buf; ; *d++ = '\\', *d++ = *s++) {
	if((k = strcspn(s, spec)))
	    do *d++ = *s++; while(--k);
	if(!*s) break;
    }
    *d = 0;
    return d - buf;
}

/*
 * enclose in braces and escape &\; or unbalanced }{ with a backslash,
 * when needed. otherwise just copy the string.
 */
void escape_command(str, buf)
char str[LINELEN], buf[LINELEN];
{
    char *d, *s, *k;
    if(!skip_arg(str, CMDSEP) && skip_arg(str, '\0')) {
	strcpy(buf, str);
	return;
    }
    d = buf; s = k = str; *d++ = '{';
    do {
	if(s == k) {
	    if(*s == CMDSEP || *s == '&' || *s == '\\' || *s == '}' ||
		    (*s == '{' && !(k = skip_arg(s + 1, '}')) && (k = s, 1))) {
		if(d - buf >= LINELEN - 2) break;
		*d++ = '\\';
	    }
	    k++;
	}
	*d++ = *s++;
    } while(d - buf < LINELEN - 1 && (*s || (*d++ = '}', 0)));
    *d = '\0';
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *d = s, c;
    do {
	c = *s++;
	if(c == '\\') c = *s++;
    } while((*d++ = c));
}

/*
 * make an escape sequence printable (for use in definition-files)
 */
void escape_seq(seq, buf)
char *seq, *buf;
{
    char *s = seq, *p = buf;
    unsigned char c;
    /*
     * rules: control chars are written as ^X, other unprintables
     * in octal \nnn form. Don't bother to escape =\^ with a \,
     * just octalize 'em.
     */
    while((c = *s++)) {
	if(c >= '\177' || c == '=' || c == '\\' || c == '^') {
	    sprintf(p, "\\%03o", (int)c);
	    p += 4;
	    continue;
	} else if(c < ' ')
	    *p++ = '^', c |= '@';
	*p++ = c;
    }
    *p = '\0';
}

/*
 * unescape, but also recognize ^c, \nnn, \x syntax.
 */
void unescape_seq(s)
char *s;
{
    char *d = s, c;
    unsigned int t;
    do switch((c = *s++)) {
	case '^':
	    c = 037 & *s++;
	    if(!c) c = '\200';
	    break;
	case '\\':
	    c = *s++;
	    t = ctoi(c);
	    if(t < 010) {
		c = t;  t = ctoi(*s);
		if(t < 010) {
		    s++; c <<= 3; c += t;  t = ctoi(*s);
		    if(t < 010) { s++; c <<= 3; c += t; }
		}
		if(!c) c = '\200';
	    }
    } while((*d++ = c));
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_arg(str, param)
char *str;
struct param param[10];
{
    int i = 1;
    char *s = str;
    param[0].s = s;
    param[0].len = strlen(s);
    do {
	param[i].s = s = skipspace(s);
	if(!*s) break;
	while(*++s && *s != ' ');
	param[i].len = s - param[i].s;
    } while(++i < 10 && *s++);
    while(i < 10) param[i++].len = 0;
}

/*
 * like strchr, but honour (possibly nested) braces and escaping backslashes.
 * return pointer to the delim in str, or NULL.
 */
static char *skip_arg(str, delim)
char *str;
char delim;
{
    char *s = str - 1;
    for( ; ; ) {
	if(*++s == delim) return s;
	if(!*s || *s == '}') break;
	if(*s != '\\') {
	    if(*s == '{') {
		int nested = 1;
		for( ; ; )	/* inner loop (text within {}) */
		    if(*++s)
			if(*s != '{') {
			    if(*s != '}') {
				if(*s == '\\' && !*++s) return 0;
			    } else if(!--nested) break;
			} else nested++;
		    else return 0;
	    }
	} else if(!*++s) {
	    if(delim) break;
	    return s;
	}
    }
    return 0;
}

int get_token(args, table)
char **args, *table[];
{
    int i;
    char *word = *args;
    *args = chop_str(word, ' ');
    for(i = 0; table[i]; i++)
	if(!strcmp(word, table[i])) break;
    return i;
}

#if 0
char *next_arg(str)
char *str;
{
    char *s = skip_arg(str, ' ');
    if(s) {
	for(*s++ = 0; *s == ' '; s++);
	if(!*s) s = 0;
    }
    return s;
}
#endif

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = skip_arg(str, delim);
    if(s) *s++ = 0;
    return s;
}

/*
 * like chop_arg, but no special handling of {\;
 * used as a replacement of strtok
 */
char *chop_str(str, delim)
char *str, delim;
{
    char *s = str;
    if(*s) while(*s != delim && *++s);
    if(*s) for(*s = 0; *++s == delim; );
    return *s ? s : 0;
}

#ifndef skipspace
/*
 * return pointer to next non-blank char
 * NOTE: only space is considered whitespace.
 * editline is all printable chars anyway.
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}
#endif

/*
 * return pointer to next non-digit
 */
char *skipnum(p)
char *p;
{
    if(*p == '-') p++;
    while(isdigit(*p)) p++;
    return p;
}
